from __future__ import annotations
import re

###########################################################
# MIDSV conversion (from CS tag to MIDSV)
###########################################################


def split_cstag(cstag: str) -> list[str]:
    """Split cstag
    Args:
        cstag (str): a string of CS tags
    Returns:
        list[str]: splitted CS tags
    Example:
        >>> from midsv import convert
        >>> cstag = "cs:Z:=ACGT*ag=C-g=T+t=ACGT"
        >>> convert.split_cstag(cstag)
        "['=ACGT', '*ag', '=C', '-g', '=T', '+t', '=ACGT']"
    """
    cstag_splitted = []
    for cs in re.split(r"(=|\*|\-|\+|\~)", cstag):
        if not cs or cs == "cs:Z:":
            continue
        if re.match(r"(=|\*|\-|\+|\~)", cs):
            cstag_splitted.append(cs)
        else:
            cstag_splitted[-1] += cs
    return cstag_splitted


def to_midsv(cstag_splitted: list[str]) -> list[str]:
    """Convert to MIDSV
    Args:
        cstag_splitted (list[str]): splitted CS tags
    Returns:
        list[str]: splitted MIDSV
    Examples:
        >>> from midsv import convert
        >>> cstag_splitted = ['=ACGT', '*ag', '=C', '-gg', '=T', '+t', '=ACGT']
        >>> convert.to_midsv(cstag_splitted)
        "['MMMM', 'S', 'M', 'DD', 'M', 'I', 'MMMM']"

        >>> cstag_splitted = ['=ACGT', '~tg5ca', '=ACGT']
        >>> convert.to_midsv(cstag_splitted)
        "['MMMM', 'DDDDD', 'MMMM']"
    """
    midsv_splitted = []
    for cs in cstag_splitted:
        num = len(cs) - 1
        if cs.startswith("="):
            midsv_splitted.append("M" * num)
        elif cs.startswith("+"):
            midsv_splitted.append("I" * num)
        elif cs.startswith("-"):
            midsv_splitted.append("D" * num)
        elif cs.startswith("~"):
            match = re.match(r"([a-z]+)([0-9]+)([a-z]+)", cs[1:])
            _, splice, _ = match.groups()
            midsv_splitted.append("D" * (int(splice)))
        else:
            midsv_splitted.append("S")
    return midsv_splitted


def numerize_insertion(midsv: list[str]) -> list[str]:
    """Convert insertion to numeric numbers
    Example:
        >>> from midsv import convert
        >>> midsv = ['MMM', 'III', 'D', 'S']
        >>> convert.numerize_insertion(midsv)
        "['MMM', 3, 'D', 'S']"
    """
    for i, m in enumerate(midsv):
        if m.startswith("I"):
            midsv[i] = len(m)
    return midsv


def slide_insertion(midsv_numerized: list[str]) -> list[str]:
    """Append one base from the next index at an inserted base.

    Args:
        midsv (list[str]): numerized MIDSV

    Returns:
        list[str]: slided MIDSV

    Example:
        >>> from midsv import convert
        >>> midsv = [3, 'M', 4, "S", "MM"]
        >>> convert.slide_insertion(midsv)
        "['3M', '4S', "MM"]"
    """
    for i, m in enumerate(midsv_numerized):
        if not m:
            continue
        if i + 1 == len(midsv_numerized):
            midsv_numerized[i] = str(midsv_numerized[i])
            continue
        if isinstance(m, int):
            midsv_numerized[i] = str(m) + midsv_numerized[i + 1][0]
            midsv_numerized[i + 1] = midsv_numerized[i + 1][1:]
    return [m for m in midsv_numerized if m]


def to_string(midsv: list[str]) -> str:
    """Convert to string in CSV format

    Args:
        midsv (list[str]): a list of MIDSV

    Returns:
        str: MIDSV

    Example:
        >>> from midsv import convert
        >>> midsv = ['3M', '4S', "MM", "DDD"]
        >>> convert.to_string(midsv)
        "3M,4S,M,M,D,D,D"
    """
    midsv_csv = []
    for m in midsv:
        if m[0].isdigit():
            midsv_csv.append(m)
            continue
        for mm in m:
            midsv_csv.append(mm)
    return ",".join(midsv_csv)


def cstag_to_midsv(cstag: str) -> str:
    """Integrated function to convert cstag to MIDSV

    Args:
        cstag (str): a long format cstag

    Returns:
        str: MIDSV format
    """
    cstag_splited = split_cstag(cstag)
    midsv_splited = to_midsv(cstag_splited)
    midsv_numerized = numerize_insertion(midsv_splited)
    midsv_slided = slide_insertion(midsv_numerized)
    return to_string(midsv_slided)


###########################################################
# Split CS tags
###########################################################


def cstag_to_cssplit(cstag: str) -> str:
    """Generate CS SPLIT, a comma-separated nucreotide sequence corresponding to MIDSV

    Args:
        cstag (str): a long format cstag

    Returns:
        str: cs split

    Examples:
        >>> cstag = "cs:Z:=A+ttt=CC-aa=T*ag=TT"
        >>> convert.cstag_to_cssplit(cstag)
        "=A,+T|+T|+T|=C,=C,-A,-A,=T,*AG,=T,=T"

        >>> cstag = "cs:Z:=A~ta10cg=T"
        >>> convert.cstag_to_cssplit(cstag)
        "=A,N,N,N,N,N,N,N,N,N,N,=T"
    """
    cstag_splitted = split_cstag(cstag)
    cssplits = []
    for i, cs in enumerate(cstag_splitted):
        if len(cs) == 1:
            continue
        op = cs[0]
        if op == "+":
            insertion = list(cs[1:])
            insertion = "+" + "|+".join(insertion)
            if i + 1 == len(cstag_splitted):
                cssplits.append(insertion)
                break
            next_cstag = cstag_splitted[i + 1]
            next_op = next_cstag[0]
            if next_op == "*":
                cssplits.append(insertion + "|" + next_cstag)
                cstag_splitted[i + 1] = "*"
            elif next_op == "~":
                insertion = insertion + "|" + "N"
                cssplits.append(insertion)
                match = re.match(r"([a-z]+)([0-9]+)([a-z]+)", next_cstag[1:])
                left, splice, right = match.groups()
                cstag_splitted[i + 1] = f"{next_op}{left}{int(splice)-1}{right}"
            else:
                insertion = insertion + "|" + next_cstag[:2]
                cssplits.append(insertion)
                cstag_splitted[i + 1] = next_op + next_cstag[2:]
        elif op == "*":
            cssplits.append(cs)
        elif op == "~":
            match = re.match(r"([a-z]+)([0-9]+)([a-z]+)", cs[1:])
            _, splice, _ = match.groups()
            cssplits.extend(["N"] * int(splice))
        else:
            cs = list(cs[1:])
            cs = op + f",{op}".join(cs)
            cssplits.append(cs)
    return ",".join([cs.upper() for cs in cssplits])


###########################################################
# Phred score
###########################################################


def ascii_to_phred(ascii: str) -> str:
    return str(ord(ascii) - 33)


def qual_to_qscore_midsv(qual: str, midsv: str) -> str:
    """Convert ascii quality to phred score.
    To adjust the same length as MIDSV, insertion is discarded and deletion is interpolated as -1.

    Args:
        qual (str): QUAL in SAM format
        midsv (str): MIDSV

    Returns:
        str: Phred quality score with indel compensation ('-1' is assigned to the deletion loci.)
    """
    qscore = []
    idx = 0
    for m in midsv.split(","):
        if m == "D":
            qscore.append("-1")
            idx -= 1
        elif m[0].isdigit():
            num_insertion = int(m[:-1])
            insertion = []
            for j in range(idx, idx + num_insertion):
                insertion.append(ascii_to_phred(qual[j]) + "|")
            if m[-1] == "D":
                insertion.append("-1")
                idx -= 1
            else:
                insertion.append(ascii_to_phred(qual[j + 1]))
            qscore.append("".join(insertion))
            idx += num_insertion
        else:
            qscore.append(ascii_to_phred(qual[idx]))
        idx += 1
    return ",".join(qscore)


def qual_to_qscore_cssplit(qual: str, cssplit: str) -> str:
    """Convert ascii quality to phred score.
    To adjust the same length as cssplit, insertion is discarded and deletion is interpolated as -1.

    Args:
        qual (str): QUAL in SAM format
        cssplit (str): CSSPLIT

    Returns:
        str: Phred quality score with indel compensation ('-1' is assigned to the deletion loci.)
    """
    qscore = []
    idx = 0
    for cs in cssplit.split(","):
        if cs.startswith("-") or cs == "N":  # N = spliced loci
            qscore.append("-1")
            idx -= 1
        elif cs.startswith("+"):
            num_insertion = len(cs.split("|")) - 1
            insertion = []
            for j in range(idx, idx + num_insertion):
                insertion.append(ascii_to_phred(qual[j]) + "|")
            if cs.split("|")[-1].startswith("-") or cs.split("|")[-1] == "N":
                insertion.append("-1")
                idx -= 1
            else:
                insertion.append(ascii_to_phred(qual[j + 1]))
            qscore.append("".join(insertion))
            idx += num_insertion
        else:
            qscore.append(ascii_to_phred(qual[idx]))
        idx += 1
    return ",".join(qscore)



from __future__ import annotations

from collections.abc import Iterator

from midsv import convert, format, proofread, validate


def transform(
    sam: list[list] | Iterator[list],
    midsv: bool = True,
    cssplit: bool = True,
    qscore: bool = True,
    keep: list[str] = None,
) -> list[dict]:
    """Integrated function to perform MIDSV conversion

    Args:
        sam (list[list]): Lists ot SAM format
        midsv (bool, optional): Output MIDSV. Defaults to True.
        cssplit (bool, optional): Output CSSPLIT. Defaults to True.
        qscore (bool, optional): Output QSCORE. Require `midsv == True` or `cssplit == True`. Defaults to True.
        keep (set(str), optional): Subset of {'FLAG', 'POS', 'SEQ', 'QUAL', 'CIGAR', 'CSTAG'} to keep. Defaults to set().
    Returns:
        list[dict]: Dictionary containing QNAME, RNAME, MIDSV, and QSCORE
    """

    if midsv or cssplit:
        pass
    else:
        raise ValueError("Either midsv or cssplit must be True")

    keep = set(keep) if keep else set()
    if keep != set() and not keep.issubset({"FLAG", "POS", "SEQ", "QUAL", "CIGAR", "CSTAG"}):
        raise ValueError("'keep' must be a subset of {'FLAG', 'POS', 'SEQ', 'QUAL', 'CIGAR', 'CSTAG'}")

    sam = list(sam)
    validate.sam_headers(sam)
    validate.sam_alignments(sam)

    sqheaders = format.extract_sqheaders(sam)
    samdict = format.dictionarize_sam(sam)

    if qscore and any(s["QUAL"] == "*" for s in samdict):
        raise ValueError("qscore must be False because the input does not contain QUAL")

    samdict = format.remove_softclips(samdict)
    samdict = format.remove_resequence(samdict)

    for alignment in samdict:
        if midsv:
            alignment["MIDSV"] = convert.cstag_to_midsv(alignment["CSTAG"])
        if cssplit:
            alignment["CSSPLIT"] = convert.cstag_to_cssplit(alignment["CSTAG"])
        if midsv and qscore:
            alignment["QSCORE"] = convert.qual_to_qscore_midsv(alignment["QUAL"], alignment["MIDSV"])
        elif cssplit and qscore:
            alignment["QSCORE"] = convert.qual_to_qscore_cssplit(alignment["QUAL"], alignment["CSSPLIT"])

    samdict_polished = proofread.join(samdict)
    samdict_polished = proofread.pad(samdict_polished, sqheaders)
    samdict_polished = proofread.remove_different_length(samdict_polished, sqheaders)
    samdict_polished = proofread.select(samdict_polished, keep)
    return samdict_polished



from __future__ import annotations
import re
from itertools import groupby
from copy import deepcopy

###########################################################
# Format headers and alignments
###########################################################


def extract_sqheaders(sam: list[list]) -> dict[str, int]:
    """Extract SN (Reference sequence name) and LN (Reference sequence length) from SQ header

    Args:
        sam (list[list]): a list of lists of SAM format

    Returns:
        dict: a dictionary containing (multiple) SN and LN
    """
    sqheaders = [s for s in sam if "@SQ" in s]
    SNLN = {}
    for sqheader in sqheaders:
        sn_ln = [sq for sq in sqheader if re.search(("SN:|LN:"), sq)]
        sn = sn_ln[0].replace("SN:", "")
        ln = sn_ln[1].replace("LN:", "")
        SNLN.update({sn: int(ln)})
    return SNLN


def dictionarize_sam(sam: list[list[str]]) -> list[dict[str | int]]:
    """Extract mapped alignments from SAM

    Args:
        sam (list[list]): a list of lists of SAM format including CS tag

    Returns:
        dict: a dictionary containing QNAME, RNAME, POS, QUAL, CSTAG and RLEN
    """
    aligns = []
    for alignment in sam:
        if alignment[0].startswith("@"):
            continue
        if alignment[2] == "*":
            continue
        if alignment[9] == "*":
            continue
        for i, a in enumerate(alignment):
            if a.startswith("cs:Z:") and not re.search(r":[0-9]+", alignment[i]):
                idx_cstag = i
        alignments = dict(
            QNAME=alignment[0].replace(",", "_"),
            FLAG=int(alignment[1]),
            RNAME=alignment[2],
            POS=int(alignment[3]),
            CIGAR=alignment[5],
            SEQ=alignment[9],
            QUAL=alignment[10],
            CSTAG=alignment[idx_cstag],
        )
        aligns.append(alignments)
    aligns = sorted(aligns, key=lambda x: [x["QNAME"], x["POS"]])
    return aligns


###########################################################
# Remove undesired reads
###########################################################


def split_cigar(cigar: str) -> list[str]:
    cigar_iter = iter(re.split(r"([MIDNSHPX=])", cigar))
    cigar_splitted = [i + op for i, op in zip(cigar_iter, cigar_iter)]
    return cigar_splitted


def remove_softclips(alignments: list[dict[str | int]]) -> list[dict[str | int]]:
    """Remove softclip information from SEQ and QUAL.

    Args:
        alignments (list[dict[str, str | int]]): disctionalized alignments

    Returns:
        list[list]: disctionalized SAM with trimmed softclips in QUAL
    """
    alignments_softclips_removed = []
    for alignment in alignments:
        cigar = alignment["CIGAR"]
        if "S" not in cigar:
            alignments_softclips_removed.append(alignment)
            continue
        cigar_split = split_cigar(cigar)
        left, right = cigar_split[0], cigar_split[-1]
        if "S" in left:
            left = int(left[:-1])
            alignment["SEQ"] = alignment["SEQ"][left:]
            alignment["QUAL"] = alignment["QUAL"][left:]
        if "S" in right:
            right = int(right[:-1])
            alignment["SEQ"] = alignment["SEQ"][:-right]
            alignment["QUAL"] = alignment["QUAL"][:-right]
        alignments_softclips_removed.append(alignment)
    return alignments_softclips_removed


def return_end_of_current_read(alignment: dict[str, str | int]) -> int:
    start_of_current_read = alignment["POS"]
    cigar = alignment["CIGAR"]
    cigar_split = split_cigar(cigar)
    alignment_length = 0
    for cig in cigar_split:
        if "M" in cig or "D" in cig or "N" in cig:
            alignment_length += int(cig[:-1])
    return start_of_current_read + alignment_length - 1


def realign_sequence(alignment: dict[str, str | int]) -> dict:
    """Discard insertion, and add deletion and spliced nucleotides to unify sequence length"""
    cigar = alignment["CIGAR"]
    cigar_split = split_cigar(cigar)
    sequence = alignment["SEQ"]
    sequence_ignored = ["N"] * alignment["POS"]
    start = 0
    for cig in cigar_split:
        if "M" in cig:
            end = start + int(cig[:-1])
            sequence_ignored.append(sequence[start:end])
            start = end
        elif "I" in cig:
            start += int(cig[:-1])
        elif any(x in cig for x in ["D", "N"]):
            sequence_ignored.append("N" * int(cig[:-1]))
    realignment = deepcopy(alignment)
    realignment["SEQ"] = "".join(sequence_ignored)
    return realignment


def remove_resequence(alignments: list[dict[str, str | int]]) -> list[dict]:
    """Remove non-microhomologic overlapped reads within the same QNAME.
    The overlapped sequences can be (1) realignments by microhomology or (2) resequence by sequencing error.
    The 'realignments' is not sequencing errors, and it preserves the same sequence.
    In contrast, the 'resequence' is a sequencing error with the following characteristics:
    (1) The shorter reads that are completely included in the longer reads
    (2) Overlapped but not the same DNA sequence
    The resequenced fragments will be discarded and the longest alignment will be retain.
    Example reads are in `tests/data/overlap/real_overlap.sam` and `tests/data/overlap/real_overlap2.sam`

    Args:
        alignments (list[dict[str, str | int]]): disctionalized alignments

    Returns:
        list[list]: disctionalized SAM with removed overlaped reads
    """
    alignments.sort(key=lambda x: x["QNAME"])
    sam_groupby = groupby(alignments, lambda x: x["QNAME"])
    sam_nonoverlapped = []
    for _, alignments in sam_groupby:
        alignments = list(alignments)
        if len(alignments) == 1:
            sam_nonoverlapped += alignments
            continue
        alignments = [realign_sequence(alignment) for alignment in alignments]
        alignments = sorted(alignments, key=lambda x: [x["POS"]])
        is_overraped = False
        end_of_previous_read = -1
        previous_read = alignments[0]["SEQ"]
        for i, alignment in enumerate(alignments):
            if i == 0:
                start_of_previous_read = alignment["POS"] - 1
                end_of_previous_read = return_end_of_current_read(alignment)
                continue
            start_of_current_read = alignment["POS"] - 1
            end_of_current_read = return_end_of_current_read(alignment)
            # (1) The shorter reads that are completely included in the longer reads
            if start_of_previous_read <= start_of_current_read and end_of_previous_read >= end_of_current_read:
                is_overraped = True
                break
            else:
                start_overlap = max(start_of_previous_read, start_of_current_read)
                end_overlap = min(end_of_previous_read, end_of_current_read)
                for prev, curr in zip(
                    previous_read[start_overlap:end_overlap], alignment["SEQ"][start_overlap:end_overlap]
                ):
                    if prev == "N" or curr == "N":
                        continue
                    # (2) Overlapped but not the same DNA sequence
                    if prev != curr:
                        is_overraped = True
                        break
            start_of_previous_read = start_of_current_read
            end_of_previous_read = return_end_of_current_read(alignment)
        if is_overraped:
            # The longest alignment will be retain
            alignment = sorted(alignments, key=lambda x: -len(x["SEQ"]))[0]
            sam_nonoverlapped.append(alignment)
        else:
            sam_nonoverlapped += alignments
    return sam_nonoverlapped



from __future__ import annotations
import json
from pathlib import Path
from typing import Generator

###########################################################
# Read sam
###########################################################


def read_sam(path_of_sam: str | Path) -> Generator[list]:
    sam = Path(path_of_sam).read_text().strip().split("\n")
    return (s.split("\t") for s in sam)


###########################################################
# Read / Write jsonl
###########################################################


def read_jsonl(filepath: str | Path) -> Generator[dict]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.JSONDecoder(strict=False).decode(line)


def write_jsonl(dicts: list[dict], filepath: str | Path):
    with open(filepath, "w") as output:
        for d in dicts:
            json.dump(d, output)
            output.write("\n")



from __future__ import annotations
import re


###########################################################
# Validate sam format
###########################################################


def sam_headers(sam: list[list]) -> None:
    """Check headers containing SN (Reference sequence name) and LN (Reference sequence length)

    Args:
        sam (list[list]): a list of lists of SAM format

    """
    sqheaders = [s for s in sam if "@SQ" in s]
    if not sqheaders:
        raise ValueError("Input does not have @SQ header")


def sam_alignments(sam: list[list]) -> None:
    """Check alignments are mapped and have long-formatted cs tag

    Args:
        sam (list[list]): a list of lists of SAM format including CS tag
    """
    is_alignment = False
    for alignment in sam:
        if alignment[0].startswith("@"):
            continue
        if alignment[2] == "*":
            continue
        is_alignment = True
        if len(alignment) < 10:
            raise ValueError("Alighment may not be SAM format")
        idx_cstag = -1
        for i, a in enumerate(alignment):
            if a.startswith("cs:Z:") and not re.search(r":[0-9]+", alignment[i]):
                idx_cstag = i
                break
        if idx_cstag == -1:
            raise ValueError("Input does not have long-formatted cs tag")
        # if "~" in alignment[idx_cstag]:
        #     raise ValueError("long-read spliced alignment are currently not supported")
    if not is_alignment:
        raise ValueError("No alignment information")



from __future__ import annotations

from copy import deepcopy
from itertools import groupby


def is_forward_strand(flag: int) -> bool:
    """
    Determines if the read is mapped to the forward strand.

    Args:
    flag (int): The SAM flag value.

    Returns:
    bool: True if the read is mapped to the forward strand, False otherwise.
    """
    # Check if the bit 4 (0-based) is NOT set
    return not (flag & 16)


def join(samdict: list[dict]) -> list[dict]:
    """Join splitted reads including large deletion or inversion.

    Args:
        samdict (list[dict]): dictionarized SAM

    Returns:
        list[dict]: SAM with joined splitted reads to single read
    """
    sam_sorted = sorted(samdict, key=lambda x: [x["QNAME"], x["POS"]])
    sam_groupby = groupby(sam_sorted, key=lambda x: x["QNAME"])
    sam_joined = []
    for *_, alignments in sam_groupby:
        alignments = list(alignments)
        if len(alignments) == 1:
            sam_joined.append(alignments[0])
            continue
        for i, alignment in enumerate(alignments):
            #########################################
            # Inversion detection
            #########################################
            # Determine the strand (strand_first) of the first read
            if i == 0:
                sam_template = deepcopy(alignment)
                if is_forward_strand(alignment["FLAG"]):
                    first_read_is_forward = True
                else:
                    first_read_is_forward = False
                continue
            # If the strand of the next read is different from strand_first, lowercase it as an Inversion.
            if is_forward_strand(alignment["FLAG"]):
                current_read_is_forward = True
            else:
                current_read_is_forward = False
            if first_read_is_forward is not current_read_is_forward:
                if "MIDSV" in alignment:
                    alignment["MIDSV"] = alignment["MIDSV"].lower()
                if "CSSPLIT" in alignment:
                    alignment["CSSPLIT"] = alignment["CSSPLIT"].lower()
            #########################################
            # Remove microhomology
            #########################################
            previous_alignment = alignments[i - 1]
            previous_end = previous_alignment["POS"] - 1
            if "MIDSV" in alignment:
                previous_end += len(previous_alignment["MIDSV"].split(","))
            else:
                previous_end += len(previous_alignment["CSSPLIT"].split(","))
            current_start = alignment["POS"] - 1
            if "CSSPLIT" in alignment:
                previous_cssplit = previous_alignment["CSSPLIT"].split(",")
                current_cssplit = alignment["CSSPLIT"].split(",")
                if "QSCORE" in alignment:
                    previous_qscore = previous_alignment["QSCORE"].split(",")
                    current_qscore = alignment["QSCORE"].split(",")
                num_microhomology = 0
                for i in range(min(len(previous_cssplit), len(current_cssplit))):
                    if previous_cssplit[-i:] == current_cssplit[:i]:
                        if "QSCORE" in alignment:
                            if previous_qscore[-i:] == current_qscore[:i]:
                                num_microhomology = i
                        else:
                            num_microhomology = i
                #########################################
                # Update CSSPLIT
                #########################################
                alignment["CSSPLIT"] = ",".join(current_cssplit[num_microhomology:])
                if "QSCORE" in alignment:
                    alignment["QSCORE"] = ",".join(current_qscore[num_microhomology:])
                if "MIDSV" in alignment:
                    current_midsv = alignment["MIDSV"].split(",")
                    alignment["MIDSV"] = ",".join(current_midsv[num_microhomology:])
                current_start += num_microhomology
                alignment["POS"] = current_start + 1
            # Fill in the gap between the first read and the next read with a D (deletion)
            gap = current_start - previous_end
            if "MIDSV" in sam_template:
                sam_template["MIDSV"] += ",D" * gap
            if "CSSPLIT" in sam_template:
                sam_template["CSSPLIT"] += ",N" * gap
            if "QSCORE" in sam_template:
                sam_template["QSCORE"] += ",-1" * gap
            # Update sam_template
            if "MIDSV" in sam_template:
                sam_template["MIDSV"] += "," + alignment["MIDSV"]
            if "CSSPLIT" in sam_template:
                sam_template["CSSPLIT"] += "," + alignment["CSSPLIT"]
            if "QSCORE" in sam_template:
                sam_template["QSCORE"] += "," + alignment["QSCORE"]
        sam_joined.append(sam_template)
    return sam_joined


def pad(samdict: list[dict], sqheaders: dict) -> list[dict]:
    """Padding left and right flanks as "=" in MIDSV, "-1" in QUAL

    Args:
        sam (list[dict]): dictionarized SAM
        sqheaders (dict): dictionary as {SQ:LN}

    Returns:
        list[dict]: dictionarized SAM with padding as "N" in MIDSV and CSSPLIT, and "-1" in QUAL
    """
    samdict_padding = []
    for alignment in samdict:
        reflength = sqheaders[alignment["RNAME"]]
        leftpad = max(0, alignment["POS"] - 1)
        if "MIDSV" in alignment:
            rightpad = reflength - (len(alignment["MIDSV"].split(",")) + leftpad)
        else:
            rightpad = reflength - (len(alignment["CSSPLIT"].split(",")) + leftpad)
        rightpad = max(0, rightpad)
        leftpad_midsv, rightpad_midsv = "N," * leftpad, ",N" * rightpad
        leftpad_cssplit, rightpad_cssplit = "N," * leftpad, ",N" * rightpad
        leftpad_qscore, rightpad_qscore = "-1," * leftpad, ",-1" * rightpad
        if "MIDSV" in alignment:
            alignment["MIDSV"] = leftpad_midsv + alignment["MIDSV"] + rightpad_midsv
        if "CSSPLIT" in alignment:
            alignment["CSSPLIT"] = leftpad_cssplit + alignment["CSSPLIT"] + rightpad_cssplit
        if "QSCORE" in alignment:
            alignment["QSCORE"] = leftpad_qscore + alignment["QSCORE"] + rightpad_qscore
        samdict_padding.append(alignment)
    return samdict_padding


def remove_different_length(samdict: list[dict], sqheaders: dict) -> list[dict]:
    """remove different sequence length of the reference

    Args:
        sam (list[dict]): dictionarized SAM
        sqheaders (dict): dictionary as {SQ:LN}

    Returns:
        list[dict]: filtered SAM by different sequence length of the reference
    """
    samdict_filtered = []
    for alignment in samdict:
        reflength = sqheaders[alignment["RNAME"]]
        if "MIDSV" in alignment:
            if len(alignment["MIDSV"].split(",")) != reflength:
                continue
        if "CSSPLIT" in alignment:
            if len(alignment["CSSPLIT"].split(",")) != reflength:
                continue
        samdict_filtered.append(alignment)
    return samdict_filtered


def select(samdict: list[dict], keep: set[str] = None) -> list[dict]:
    """Select QNAME, RNAME, MIDSV, CSSPLIT and QSCORE

    Args:
        samdict (list[dict]): dictionarized SAM
        keep (set(str), optional): Subset of {'FLAG', 'POS', 'SEQ', 'QUAL', 'CIGAR', 'CSTAG'} to keep. Defaults to set().
    Returns:
        list[dict]: dictionarized SAM of QNAME, RNAME, MIDSV, CSSPLIT and QSCORE
    """
    keep = set(keep) if keep else set()
    selected = []
    for m in samdict:
        delkeys = {"FLAG", "POS", "SEQ", "QUAL", "CIGAR", "CSTAG"} - keep
        for key in delkeys:
            m.pop(key)
        selected.append(m)
    return selected
